#include <winsock2.h>
#include <iphlpapi.h>
#include <stdio.h>
#include <stdlib.h>

#include "MachineId.h"
#include "Logger.h"

// Link with Iphlpapi.lib
#pragma comment(lib, "IPHLPAPI.lib")

#define WORKING_BUFFER_SIZE 15000
#define MAX_TRIES 3

#define MALLOC(x) HeapAlloc(GetProcessHeap(), 0, (x))
#define FREE(x) HeapFree(GetProcessHeap(), 0, (x))


static const char* ENV_NAME = "PARASOFT_SUPPORT_NET_INTERFACES";
static const char* ENV_DEBUG = ::getenv("PARASOFT_DEBUG_NET_INTERFACES");
static const char* ENV_DEBUG_FILE = ::getenv("PARASOFT_DEBUG_NET_INTERFACES_FILE");
static PIP_ADAPTER_ADDRESSES SELECTED_ADAPTER = NULL;
static const bool IS_DEBUG_FILE = ENV_DEBUG_FILE != NULL && strcmp(ENV_DEBUG_FILE, "") != 0;
static const bool DEBUG = IS_DEBUG_FILE || (ENV_DEBUG != NULL && strcmp(ENV_DEBUG, "true") == 0);

PIP_ADAPTER_ADDRESSES findAdapter(PIP_ADAPTER_ADDRESSES pAddresses, DWORD type);
int generateMachineId(unsigned char PhysicalAddress[], int PhysicalAddressLength);
static void printDebug(const char* format, ...);
bool addressExist(char * envAdress, PIP_ADAPTER_ADDRESSES pAddress);
int convertToDecimal(char x, char y);
int charToInt(char x);
PIP_ADAPTER_ADDRESSES scanAdapters(PIP_ADAPTER_ADDRESSES pCurrAddresses, PIP_ADAPTER_ADDRESSES pAddresses);

MachineId::MachineId(void)
{
}

MachineId::~MachineId(void)
{
}

// See: http://msdn.microsoft.com/en-us/library/windows/desktop/aa365915%28v=vs.85%29.aspx
int MachineId::getMachineId()
{

	char* envInterface = ::getenv(ENV_NAME);

    DWORD dwSize = 0;
    DWORD dwRetVal = 0;

    unsigned int i = 0;

    // Set the flags to pass to GetAdaptersAddresses
    ULONG flags = GAA_FLAG_INCLUDE_PREFIX;

    // Unspecified address family (both). Other values: AF_INET, AF_INET6
    ULONG family = AF_UNSPEC;

    LPVOID lpMsgBuf = NULL;

    PIP_ADAPTER_ADDRESSES pAddresses = NULL;
    ULONG outBufLen = 0;
    ULONG Iterations = 0;

    PIP_ADAPTER_ADDRESSES pCurrAddresses = NULL;
    PIP_ADAPTER_UNICAST_ADDRESS pUnicast = NULL;
    PIP_ADAPTER_ANYCAST_ADDRESS pAnycast = NULL;
    PIP_ADAPTER_MULTICAST_ADDRESS pMulticast = NULL;
    IP_ADAPTER_DNS_SERVER_ADDRESS *pDnServer = NULL;
    IP_ADAPTER_PREFIX *pPrefix = NULL;
    // Allocate a 15 KB buffer to start with.
    outBufLen = WORKING_BUFFER_SIZE;
	
	if (envInterface != 0) {
		printDebug("[DEBUG] Checking physical address from environment variable \n");
		pAddresses = (IP_ADAPTER_ADDRESSES *)MALLOC(outBufLen);
		if (pAddresses == NULL) {
			return 0x0;
		}
		dwRetVal = GetAdaptersAddresses(family, flags, NULL, pAddresses, &outBufLen);
		if (!addressExist(envInterface, pAddresses)) {
			printDebug("[WARN] The network card with physical address set in PARASOFT_SUPPORT_NET_INTERFACE variable has not been found. Ensure the variable is properly set \n");
			pCurrAddresses = scanAdapters(pCurrAddresses, pAddresses);
			return generateMachineId(pCurrAddresses->PhysicalAddress, (int)pCurrAddresses->PhysicalAddressLength);
		}
		int machineId = 0;
		if (SELECTED_ADAPTER != NULL) {
			machineId = generateMachineId(SELECTED_ADAPTER->PhysicalAddress, (int)SELECTED_ADAPTER->PhysicalAddressLength);
		}
		printDebug("[SELECTED] Physical Address: ");
		printDebug(envInterface);
		printDebug(" (user-provided) \n");
		return machineId;
	}
	
    do {
        pAddresses = (IP_ADAPTER_ADDRESSES *) MALLOC(outBufLen);
        if (pAddresses == NULL) {
            setErrorMessage("Memory allocation failed for IP_ADAPTER_ADDRESSES struct");
            Logger::logln("Memory allocation failed for IP_ADAPTER_ADDRESSES struct");
            return 0x0;
        }
        
        // The order in which adapters appear in the list returned by this function 
        // can be controlled from the Network Connections folder: select the Advanced Settings menu item from the Advanced menu
        dwRetVal = GetAdaptersAddresses(family, flags, NULL, pAddresses, &outBufLen);
        if (dwRetVal == ERROR_BUFFER_OVERFLOW) {
            FREE(pAddresses);
            pAddresses = NULL;
        } else {
            break;
        }
        Iterations++;
    } while ((dwRetVal == ERROR_BUFFER_OVERFLOW) && (Iterations < MAX_TRIES));

    int machineId = 0;
    if (dwRetVal == NO_ERROR) {
        
        pCurrAddresses = scanAdapters(pCurrAddresses, pAddresses);
        if (pCurrAddresses) {
            machineId = generateMachineId(pCurrAddresses->PhysicalAddress, (int) pCurrAddresses->PhysicalAddressLength);
        } else {
            Logger::logln("Cannot find network adapter. Run ipconfig /all to list known adapters.");
        }

    } else {
        char messageBuffer [100];
        setErrorMessage(messageBuffer);
        Logger::logln(messageBuffer);
    }

    if (pAddresses) {
        FREE(pAddresses);
    }

    return machineId;
}

PIP_ADAPTER_ADDRESSES findAdapter(PIP_ADAPTER_ADDRESSES pAddresses, DWORD type)
{
    PWCHAR pName = NULL;
    PIP_ADAPTER_ADDRESSES pAddress = NULL;
    PIP_ADAPTER_ADDRESSES pCurrAddresses = pAddresses;
	
    while (pCurrAddresses) {
    	unsigned char * MAC = pCurrAddresses->PhysicalAddress;
    	printDebug("[DEBUG] Physical Address: ");
		printDebug("%x-%x-%x-%x-%x-%x", MAC[0], MAC[1], MAC[2], MAC[3], MAC[4], MAC[5]);
		printDebug("\t");
		printDebug("Interface type: %d", pCurrAddresses->IfType);
		printDebug("\n");
        if ((type == -1) || (pCurrAddresses->IfType == type)) {
            // Sort adapters by name (pr108045)
            if ((pCurrAddresses->PhysicalAddressLength != 0) && (pCurrAddresses->FriendlyName != NULL)) {
                if ((pName == NULL) || wcscmp(pCurrAddresses->FriendlyName, pName) < 0) {
                    pName = pCurrAddresses->FriendlyName;
                    pAddress = pCurrAddresses;
                }
            }
        }
        pCurrAddresses = pCurrAddresses->Next;
    }
    if (!pAddress) {
		printDebug("No adapters available");
		return 0x0;
	}
    unsigned char * SELECTED = pAddress->PhysicalAddress;
	printDebug("[SELECTED] ");
	printDebug("%x-%x-%x-%x-%x-%x", SELECTED[0], SELECTED[1], SELECTED[2], SELECTED[3], SELECTED[4], SELECTED[5]);
	printDebug("\t");
	printDebug("Interface type: %d", pAddress->IfType);
	printDebug("\n");
    return pAddress;
}

int generateMachineId(unsigned char PhysicalAddress[], int PhysicalAddressLength)
{
    // Copied from Linux machine id calculation
    union U {
        unsigned char c[4];
        unsigned int  n;
    } u;
    
    unsigned int cSum = 0;
    for (int i = 0; i < PhysicalAddressLength; i++) {
        cSum += PhysicalAddress[i];
    }
    if (PhysicalAddressLength > 1)
        u.c[0] = PhysicalAddress[1];
    if (PhysicalAddressLength > 3)
        u.c[1] = PhysicalAddress[3];
    if (PhysicalAddressLength > 5)
        u.c[2] = PhysicalAddress[5]; 
    u.c[3] = (unsigned char)cSum;
    return u.n; 
}

static void printDebug(const char* format, ...)
{
	if (!DEBUG) {
		return;
	}
	
    va_list argptr;
    va_start(argptr, format);
    
	if (IS_DEBUG_FILE) {
		FILE * debugFile = fopen(ENV_DEBUG_FILE, "a");
		if (debugFile != NULL) {
		    vfprintf(debugFile, format, argptr);
			fclose(debugFile);
		} else {
		    printf("[Wrong PARASOFT_DEBUG_NET_INTERFACES_FILE] %s\n", ENV_DEBUG_FILE);
		    vfprintf(stdout, format, argptr);
		}
	} else {
	    vfprintf(stdout, format, argptr);
	}
	
    va_end(argptr);
    
}

bool addressExist(char * envAdress, PIP_ADAPTER_ADDRESSES pAddress) {
	int notMatchedAdapters = 0;
	int adapters = 0;
	int index = 0;
	while (pAddress){
		unsigned char * macAddress = pAddress->PhysicalAddress;
		int pAddressLength = pAddress->PhysicalAddressLength;
		for (int i = 0; i < pAddressLength; i++) {
			unsigned char octet = macAddress[i];
			char x = envAdress[index];
			index++;
			char y = envAdress[index];
			index++;
			if (envAdress[index] == ':' || envAdress[index] == '-') {
				index++;
			}
			if (octet != convertToDecimal(x,y)) {
				index = 0;
				break;
			}
		}
		if (index != 0) {
			SELECTED_ADAPTER = pAddress;
			return true;
		}

		pAddress = pAddress->Next;
	}
	return false;
}

int convertToDecimal(char x, char y)
{
	int count1 = charToInt(x);
	int count2 = charToInt(y);
	
	return (count1 * 16) + count2;
}

int charToInt(char x)
{
	if (isdigit(x)) {
		return x - '0';
	}
	char hexadecimalChars[16] = { '0', '1', '2', '3', '4', '5', '6', '7', '8', '9', 'A', 'B', 'C', 'D', 'E', 'F' };

	for (int i = 0; i < sizeof(hexadecimalChars); i++) {
		if (x == hexadecimalChars[i]) {
			return i;
		}
	}
	return x;
}

PIP_ADAPTER_ADDRESSES scanAdapters(PIP_ADAPTER_ADDRESSES pCurrAddresses, PIP_ADAPTER_ADDRESSES pAddresses)
{
	// Ethernet card first
        printDebug("[DEBUG] Scanning ethernet adapters \n");
        pCurrAddresses = findAdapter(pAddresses, IF_TYPE_ETHERNET_CSMACD);
        
        // Wireless card next
        if (!pCurrAddresses) {
        	printDebug("[DEBUG] Scanning wireless adapters \n");
            pCurrAddresses = findAdapter(pAddresses, IF_TYPE_IEEE80211);
        }
        
        // Other network card
        if (!pCurrAddresses) {
        	printDebug("[Scanning] other adapters \n");
            pCurrAddresses = findAdapter(pAddresses, -1);
        }
        return pCurrAddresses;
        
}